import matplotlib.pyplot as plt
import numpy as np
from functools import reduce
from pathlib import Path
import json
import math
import random
from typing import *


def random_color() -> list:
    """
    随机生成十六进制颜色值
    :return:
    """
    color = ['#' + ''.join([random.choice('0123456789ABCDEF') for _ in range(6)])]
    return color

class DrawPlot(object):
    def __init__(self, sys_name, save_path=None):
        self.sys_name = sys_name
        self.sava_path = save_path
        self.red = '#e60000'
        self.green = '#2eb82e'
        self.blue = '#1a52ff'
        plt.rcParams['font.sans-serif'] = ['SimHei']  # 替换全局sans-serif字体(黑体)
        plt.rcParams['axes.unicode_minus'] = False  # 解决坐标轴负号显示问题

    def draw_pie(self, title=None, color: list = None, show_plot: bool = False, **kwargs) -> str:
        """
        绘制饼图的方法，传入任意键值对参数，值都为int类型，如 成功=10 或者 success=10
        :param title: 图像标题，不填写为默认值
        :param color: 颜色列表，对应后面的数据，不填写默认绿色、红色、蓝色轮换
        :param show_plot: 是否立即显示图像，默认False
        :param kwargs: 键值对，值为int类型
        :return: 如果设置保存地址则返回图片路径，否则返回None
        """
        if color is None:
            color = [self.green, self.red, self.blue]
        # pie_label = list(kwargs.keys())
        # pie_val = list(kwargs.values())
        pie_data = [x for x in list(kwargs.items()) if x[1] > 0]
        pie_label = [x[0] for x in pie_data]
        pie_val = [x[1] for x in pie_data]
        plt.pie(pie_val, labels=pie_label, colors=color, autopct='%.0f%%', textprops={"fontsize": 20})
        plt.axis("equal")
        if title is None:
            title = f"{self.sys_name}系统通过率饼状图"
        plt.title(title)
        pic_path = None
        if self.sava_path is not None:
            pic_path = self.sava_path + '/' + title + '.png'
            plt.savefig(pic_path)
        if show_plot is not False:
            plt.show()
        plt.close()
        return pic_path

    def draw_bar(self, title=None, color: list = None, show_plot: bool = False, labels: list = None, **kwargs) -> str:
        """
        绘制柱状图的方法
        :param title: 图像的title名称，填写则保存图片也以此为准
        :param color: 颜色，列表形式，如果颜色数量与小柱状的数量不对应，则循环获取颜色
        :param show_plot: 布尔值，是否立即显示图像，默认False
        :param labels: 列表，当传入数据的是列表形式，则需要填写labels，对应数据的标签
        :param kwargs: 测试数据，键值对方式传入，模块:数据，数据可以是字典或者列表，如果是嵌套列表，需要填写labels参数
        :return: 填写保存路径则返回图片保存路径，否则返回None
        """
        # 数据形式：
        # {"模块1": {"success": 10, "failed": 6, "skip": 8}, "模块2": {"success": 20, "failed": 8, "skip": 6}}
        # {"模块1": [10, 6, 8], "模块2": [20, 8, 6]}, labels=["成功", "失败", "跳过"]
        bar_width = 0.1
        if color is None:
            color = [self.green, self.red, self.blue]
        model_labels = list(kwargs.keys())
        x = np.arange(len(model_labels))
        if type(kwargs[model_labels[0]]) is dict:
            cases = list(kwargs.values())
            labels = [list(x.keys()) for x in list(kwargs.values())][0]
        elif type(kwargs[model_labels[0]]) is list or tuple:
            if labels is None:
                raise ValueError("输入未包含labels，列表形式请传入等量的labels")
            cases = [dict(zip(labels, x)) for x in list(kwargs.values())]
        else:
            raise ValueError("输入类型只能是字典、列表、元组")
        if len(cases) > 1:
            case_count = list(zip(*[list(z.values()) for z in cases]))
        else:
            case_count = list(cases[0].values())
        # print(cases)
        add_width = 0
        color_index = 0
        # print("case_count: ", case_count)
        y_max = np.array(case_count).max()
        # print('y_max: ', y_max)
        plt.ylim(0, y_max + 1)
        for index, item in enumerate(case_count):
            rects = plt.bar(x + bar_width * add_width, item, width=bar_width, color=color[color_index],
                            label=labels[index])
            for rect in rects:
                height = rect.get_height()
                plt.text(rect.get_x() + rect.get_width() / 2, height, str(height), ha="center", va="bottom", size=13)
            add_width += 1
            if color_index == len(color) - 1:
                color_index = 0
            else:
                color_index += 1
        if title is None:
            title = f"{self.sys_name}系统模块通过率柱状图"
        plt.title(title)
        plt.xticks(x + bar_width * (add_width - 1) / 2, model_labels)
        pic_path = None
        plt.legend(loc='best')
        if self.sava_path is not None:
            pic_path = self.sava_path + "/" + title + ".png"
            plt.savefig(pic_path)
        if show_plot is not False:
            plt.show()
        plt.close()
        return pic_path

    def allure_data(self, result_path) -> Tuple[dict, list, dict]:
        """
        获取allure报告中用例运行数据，allure中应有模块、和用例，即json文件中，需呀有name值代表模块，steps代表各个用例
        :param result_path: 报告数据文件保存文件夹路径
        :return: model_pass_rate 模块通过率字典dict{list[list]}形式，每个小list代表一个用例运行数据，可直接解包成为柱状图的数据
                 status_name 用例的状态类型，可用在柱状图的labels中
                 model_runtime 每个模块的运行时间dict{dict}，内部的字典代表各个用例运行耗时，可直接解包用作折线图的数据
        """
        json_files = Path(result_path).glob('*.json')       # 获取路径下所有的json文件
        cases = []
        pass_rate = []
        run_time_consum = []
        case_titles = []
        for json_path in json_files:                # 获取所有文件，只取第一个key有"name"的值
            with open(json_path, 'r', encoding='utf8') as f:
                data = json.loads(f.read())         # 转换字典形式
                if "name" in list(data.keys()):
                    cases.append(data)
        model_list = [x["name"] for x in cases]         # 获取所有的"name"，即为模块名称
        steps = [x["steps"] for x in cases]             # 获取所有的步骤，列表的形式
        for step in steps:                              # 循环所有步骤
            status = [x["status"] for x in step]        # 获取步骤中的status值，即为用例结果
            pass_rate.append(status)                    # 添加到列表中备用
            run_time = [round((x['stop']-x['start'])/1000, 3) for x in step]        # 获取step的运行时间，保留3位小数
            run_time_consum.append(run_time)            # 在run_time_consum列表中暂存
            case_title = [x['name'] for x in step]      # 获取所有steps的"name"，即为用例的名称
            case_titles.append(case_title)              # 暂存在case_titles中
        dc_list = []
        for stu in pass_rate:                   # 所有的状态，获取其在列表中出现的次数
            dc = {}
            for item in stu:
                dc[item] = dc.get(item, 0) + 1  # 字典获取值，如果没有值，则默认第一个值为0
            dc_list.append(dc)              # 暂存在dc_list列表中
        # 获取所有的状态名称，进行重复过滤
        status_name = list(
            set(reduce(lambda x, y: list(x.keys()) + list(y.keys()) if type(x) is dict else x + list(y.keys()),
                       dc_list)))
        all_status = []
        for dc_item in dc_list:         # 循环所有状态列表，多个数据修改成数量一样的状态
            status_model = []
            for key in status_name:
                status_model.append(dc_item.get(key, 0))
            all_status.append(status_model)
        model_pass_rate = dict(zip(model_list, all_status))     # 模块通过率，组合成字典的形式，作为数据返回
        case_runtime = [dict(zip(*x)) for x in list(zip(case_titles, run_time_consum))]
        model_runtime = dict(zip(model_list, case_runtime))     # 模块运行时间，包含每个用例运行时间
        return model_pass_rate, status_name, model_runtime

    def draw_plot(self, show_plot: bool = False, **kwargs):
        """
        绘制模块和用例耗时的折线图
        :param show_plot: 是否立即显示图像
        :param kwargs: 耗时的数据大致数据形式为- {"模块1": {“用例1”: 2.54, "用例2": 1.2}, "模块2": {"用例1": 3.12}}
        :return: 当设置了save_path，返回保存的文件路径，否则返回None
        """
        # models = list(kwargs.keys())
        models = [x for x in list(kwargs.keys()) if type(kwargs[x]) is dict]
        model_length = len(models)
        case_labels = [list(x.keys()) for x in list(kwargs.values()) if type(x) is dict]
        case_runtime_list = [list(x.values()) for x in list(kwargs.values()) if type(x) is dict]
        enum_case_time = list(zip(case_labels, case_runtime_list))
        fig = plt.figure(figsize=(10, 10))
        grid = plt.GridSpec(math.ceil(model_length/2 + 1), 2, wspace=0.2)       # math进一法取整，设置grid栅栏分布
        # 模块运行总时间折线图
        ax = fig.add_subplot(grid[0, 0:2])
        sum_runtime = [round(np.array(x).sum(), 3) for x in case_runtime_list]        # 同级各个模块用例耗时总和
        y_max = math.ceil(np.array(sum_runtime).max())                                 # 获取最大值
        ax.set_ylim(0, y_max + 1)             # 设置y轴值
        ax.plot(models, sum_runtime, 'o-', color=self.red)                  # 绘制折线图
        for x, y in zip(models, sum_runtime):                               # 折线图折点添加文字
            ax.text(x, y+0.1, y, ha="center", va="bottom", size=12)
        ax.set_title(f"{self.sys_name}系统模块运行耗时")
        ax.set_xlabel("模块")
        ax.set_ylabel("运行时间(s)")
        # 各个模块用例运行时间
        plot_row = 1
        plot_col = 0
        for index, case in enumerate(enum_case_time):
            case_y = math.ceil(np.array(case[1]).max())                            # 获取所有用例的耗时最大值
            ax = fig.add_subplot(grid[plot_row, plot_col])
            ax.set_ylim(0, case_y + 1)
            ax.plot(['\n'.join(c) for c in case[0]], case[1], 'o-', color=self.red)
            for x, y in zip(case[0], case[1]):
                ax.text(x, y+0.1, y, ha="center", va="bottom", size=12)
            ax.set_title(f"{models[index]}-模块用例运行时间")
            ax.set_xlabel("用例")
            ax.set_ylabel("运行时间(s)")
            if plot_col == 1:
                plot_col = 0
                plot_row += 1
            else:
                plot_col += 1
        pic_path = None
        if self.sava_path is not None:
            pic_path = self.sava_path + '/' + f'{self.sys_name}系统运行时间折线图.png'
            plt.savefig(pic_path)
        if show_plot is not False:
            plt.show()
        plt.close()
        return pic_path


import time
import traceback
from functools import wraps, reduce
from XLRP.drawplot import DrawPlot
from pathlib import Path
import os
from XLRP.xlconf import XlConf
import openpyxl
from openpyxl.styles import Font, PatternFill, Alignment
from openpyxl.utils import get_column_letter


_test_result = {"sys": '', "model": [], 'step': []}
_error = {}

class SysName:
    """
    获取系统名称的类
    """
    def __init__(self, sys_name):
        self.sys_name = sys_name

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        _test_result["sys"] = self.sys_name

    def __call__(self, cls):
        _test_result['sys'] = self.sys_name

        @wraps(cls)
        def decorator(*args, **kwargs):
            res = cls(*args, **kwargs)
            return res
        return decorator

class ModelName:
    """
    获取模块名称类
    """
    def __init__(self, model_name):
        self.model = model_name

    def __enter__(self):
        ...

    def __exit__(self, exc_type, exc_val, exc_tb):
        _test_result['model'].append(self.model)

    def __call__(self, func):
        @wraps(func)
        def decorator(*args, **kwargs):
            _test_result["model"].append(self.model)
            res = func(*args, **kwargs)
            return res
        return decorator

class StepName:
    """
    获取step步骤(用例)的类
    """
    def __init__(self, step_name):
        self.step = step_name

    def __enter__(self):
        self.start_time = time.time()

    def __exit__(self, exc_type, exc_val, exc_tb):
        if any([exc_type, exc_val, exc_tb]):
            # error_str = str([''.join(x) for x in traceback.format_exception(exc_type, exc_val, exc_tb)][0])
            error_str = reduce(lambda x, y: x + y, traceback.format_exception(exc_type, exc_val, exc_tb))
            _error[self.step] = error_str
            if exc_type is AssertionError:
                case_status = False
            else:
                case_status = 'Error'
        else:
            case_status = True
            error_str = None
        stop_time = time.time()
        run_time = round(stop_time - self.start_time, 3)
        _test_result["step"].append({self.step: {'run_time': run_time, 'status': case_status, 'msg': error_str}})

    def __call__(self, func):
        step = self.step

        @wraps(func)
        def decorator(*args, **kwargs):
            start_time = time.time()
            try:
                res = func(*args, **kwargs)
                if type(res) is bool:
                    case_status = res
                else:
                    case_status = True
                error_str = None
            except Exception as e:
                case_status = False
                res = None
                # error_str = str([''.join(x) for x in traceback.format_exc()][0])
                error_str = reduce(lambda x, y: x + y, traceback.format_exc())
            stop_time = time.time()
            run_time = round(stop_time - start_time, 3)
            _test_result["step"].append({step: {"run_time": run_time, "status": case_status, 'msg': error_str}})
            return res
        return decorator

class _ProcessingData:
    def __init__(self, data_list: dict):
        self.data_list = data_list
        self.model_list = self.data_list['model']
        self.filter_models = []
        [self.filter_models.append(x) for x in self.model_list if x not in self.filter_models]
        self.steps = self.data_list['step']
        self.zips_data = list(zip(self.model_list, self.steps))

    def pie_data(self):
        """
        修整饼图数据
        :return:
        """
        passed = 0
        failed = 0
        error = 0
        for dc in self.steps:
            status = list(dc.values())[0]['status']
            if status is True:
                passed += 1
            elif status == 'Error':
                error += 1
            elif status is False:
                failed += 1
            else:
                failed += 1
        return passed, failed, error

    def bar_data(self):
        """
        修整柱状图的测试数据
        :return:
        """
        dc_data = {}
        for model in self.filter_models:
            dc_data[model] = {'passed': 0, 'failed': 0, 'error': 0}
            for case in self.zips_data:
                if model == case[0]:
                    status = list(case[1].values())[0]['status']
                    if status is True:
                        dc_data[model]['passed'] += 1
                    elif status is False:
                        dc_data[model]['failed'] += 1
                    elif status == 'Error':
                        dc_data[model]['error'] += 1
                    else:
                        dc_data[model]['failed'] += 1
        return dc_data

    def plot_data(self):
        """
        修正绘制折线图的测试数据
        :return:
        """
        dc_data = {}
        for model in self.filter_models:
            dc_data[model] = {}
            for case in self.zips_data:
                if model == case[0]:
                    run_time = {list(case[1].keys())[0]: list(case[1].values())[0]["run_time"]}
                    dc_data[model].update(run_time)
        return dc_data

class Runner:
    def __init__(self, show_plot=False):
        self.show_plot = show_plot
        self.save_path = Path().cwd() / '.xlrp_plot'
        self.save_path.mkdir(exist_ok=True)
        self.save_path = os.fspath(self.save_path)

    def run(self, obj, param_iter=None):
        """
        运行单个用例的方法
        :param obj: 用例函数或者方法名
        :param param_iter: 参数，使用同parameter，需要列表或者元组形式，会循环传入到函数或者方法中
        :return: self
        """
        if type(param_iter) not in (list, tuple):
            if param_iter is None:
                try:
                    obj()
                except:
                    ...
            else:
                raise ValueError("The parameters must be lists or tuples")
        else:
            for param in param_iter:
                try:
                    if type(param) in (list, tuple):
                        obj(*param)
                    elif type(param) is dict:
                        obj(**param)
                    else:
                        obj()
                except:
                    ...
        return self

    def run_class(self, test_class):
        """
        运行整个Class方法，仅收集包含'xl'开头的方法
        :param test_class: 需要进行运行的类名
        :return: self
        """
        class_dirs = test_class().__dir__()
        print(class_dirs)
        func_list = [x for x in class_dirs if x.startswith("xl")]
        for func in func_list:
            print(func)
            try:
                eval(f'test_class().{func}()')
            except:
                ...
        return self

    def plot_save_excel(self, file_path=None, width_list=(500, 500, 500)):
        """
        将plot图片保存到excel的代码
        :param file_path: excel文件路径
        :param width_list: 图片的宽度(默认第三个数据，会在原基础上X2，改动在xlconf文件中)
        :return: self
        """
        sys_name = _test_result.get('sys')
        prd = _ProcessingData(_test_result)
        pie_data = prd.pie_data()
        bar_data = prd.bar_data()
        plot_data = prd.plot_data()
        dp = DrawPlot(sys_name=sys_name, save_path=self.save_path)
        pie_path = dp.draw_pie(通过=pie_data[0], 失败=pie_data[1], 错误=pie_data[2], show_plot=self.show_plot)
        bar_path = dp.draw_bar(**bar_data, show_plot=self.show_plot)
        plot_path = dp.draw_plot(**plot_data, show_plot=self.show_plot)
        pic_list = [pie_path, bar_path, plot_path]
        if file_path is not None:
            # dp.plot_to_excel(file_path=file_path, pic_list=pic_list, width_list=width_list, cells=cells)
            xc = XlConf(filepath=file_path, sheetname='TestReport', pic_list=pic_list, pic_width=width_list)
            xc.set_cell_size(create_new=True)
            del xc
            xc2 = XlConf(filepath=file_path, sheetname='TestReport', pic_list=pic_list, pic_width=width_list)
            xc2.save_plot()

    def save_default(self):
        sys_name = _test_result.get('sys')
        excel_name = f'{sys_name}系统测试EXCEL报告.xlsx'
        report_path = Path().cwd() / 'XlrpReport'
        report_path.mkdir(exist_ok=True)
        model_list = _test_result['model']
        steps = _test_result['step']
        # 创建excel表格存入数据
        wb = openpyxl.Workbook()
        ws = wb.create_sheet('TestCases')
        title_font = Font(size=16, name='微软雅黑', bold=True)
        label_font = Font(size=14, name='微软雅黑')
        content_font = Font(size=11, name='微软雅黑', bold=True)
        false_font = Font(size=11, name='微软雅黑', color='cc3333', bold=True)
        title_fill = PatternFill(fgColor='adc2eb', fill_type='solid')
        label_fill = PatternFill(fgColor='b3b3b3', fill_type='solid')
        # false_fill = PatternFill(fgColor='cc3333', fill_type='solid')
        cell_align = Alignment(horizontal='center', vertical='center', wrap_text=True)
        ws.merge_cells('A1:F1')
        ws['A1'].value = f'{sys_name}系统运行用例情况表'
        ws['A1'].font = title_font
        ws['A1'].fill = title_fill
        ws['A1'].alignment = cell_align
        labels = ['编号', '模块', '用例名称', '运行时间', '运行结果', '异常信息']
        for index, label in enumerate(labels):
            label_cell = ws.cell(row=2, column=index + 1)
            label_cell.value = label
            label_cell.font = label_font
            label_cell.fill = label_fill
            label_cell.alignment = cell_align
        for index, case_data in enumerate(zip(model_list, steps)):
            model = case_data[0]
            step = case_data[1]
            case_name = list(step.keys())[0]
            runtime = step[case_name]['run_time']
            status_origin = step[case_name]['status']
            if status_origin is True:
                status = 'PASSED'
            elif status_origin is False:
                status = 'FAILED'
            else:
                status = 'ERROR'
            msg = step[case_name]['msg']
            data = [str(index + 1), model, case_name, runtime, status, msg]
            for col, param in enumerate(data):
                case_cell = ws.cell(row=index + 3, column=col + 1)
                case_cell.value = param
                if status != 'PASSED':
                    case_cell.font = false_font
                else:
                    case_cell.font = content_font
                case_cell.alignment = cell_align
                # if status is not True:
                #     case_cell.fill = false_fill
        sava_path = os.fspath(report_path) + '/' + excel_name
        wb.save(sava_path)
        wb.close()
        del wb
        wb2 = openpyxl.load_workbook(sava_path)
        ws2 = wb2['TestCases']
        try:
            wb2.remove(wb2['Sheet'])
        except:
            ...
        for col in range(1, len(labels) + 1):
            width = 25
            if col == len(labels):
                width = 50
            ws2.column_dimensions[get_column_letter(col)].width = width
        old_model = ''
        for row in range(3, ws2.max_row + 1):
            current_model = ws2.cell(row=row, column=2).value
            if old_model == current_model:
                ws2.merge_cells(f'B{row - 1}:B{row}')
            else:
                old_model = current_model
        wb2.save(sava_path)
        wb2.close()
        del wb2
        self.plot_save_excel(file_path=sava_path)

@SysName("测试模块")
class ExcelTest:
    @ModelName('模块1')
    def xl_pr(self, a, b, step):
        with StepName(step):
            assert a == b

    @ModelName('模块2')
    def xl_pr2(self, a, b, step):
        with StepName(step):
            assert a > b/a

    @ModelName("模块3")
    @StepName("步骤")
    def xl_pr3(self):
        print(123)

    @ModelName("模块4")
    @StepName("步骤")
    def xl_pr4(self):
        print(666)

if __name__ == '__main__':
    # run(ExcelTest().pr, [[1, 2, "步骤1"], [3, 3, "步骤2"], [5, 1, "步骤3"]], show_plot=True)
    # run(ExcelTest().pr2, [[1, 2, "步骤1"], [3, 3, "步骤2"], [0, 4, "步骤3"]], show_plot=True)
    # run(ExcelTest().pr3, show_plot=True)
    runner = Runner()
    runner.run(ExcelTest().xl_pr, [[1, 2, "步骤1"], [3, 3, "步骤2"], [5, 1, "步骤3"]])
    runner.run(ExcelTest().xl_pr2, [[1, 2, "步骤1"], [3, 3, "步骤2"], [0, 4, "步骤3"]])
    runner.run(ExcelTest().xl_pr3)
    runner.save_default()
    print(_test_result)
    # runner.plot_save_excel('./test.xlsx')
    # runner.run_class(ExcelTest)
    # runner.plot_save_excel('./test.xlsx')